#!/usr/bin/perl -w

#########################################################################
# Simple PostgreSQL MVCC testing harness
# Written by Martijn van Oosterhout (C) Aug 2005
# kleptog@svana.org
# http://svana.org/kleptog/pgsql/index2.html
#
# This copied is provided under the BSD licence, see COPYING for details.
#########################################################################

use Pg;
use strict;

use constant TIMEOUT => 5;
use constant RESOLUTION => 0.1;

use constant DEFAULT_DB => "test";

my $PG_CONNECT_STRING = defined($ENV{PGDATABASE})?"":("dbname=".DEFAULT_DB);

my @connections;
my %GlobalVars;

# Set autoflush on
select STDOUT;
$|=1;

main();
exit 0;

# Main program
sub main
{
  my ($conn, $optshash, $query);
  my @comments;         # Need to remember them so they appear at the right point in the output

  while(<>)
  {
    chomp;
    if( /^\s*#/ )
    {
      push @comments, "$_\n";
      next;
    }

    next if /^\s*$/;         # Skip blank lines

    # Line beginning with space is continuation of previous query
    if( /^\s+/ )
    {
      next if not defined $query;   # Blank lines at beginning of file ok
      $query .= "\n".$_;            # \n for server side comments
      next;
    }

    if( defined $query )
    {
      my $connhash = GetConnection( $conn );

      if( $query =~ /^\s*wait\s*/i )   # This is a wait command
      {
        if( $connhash->{pending} )
        {
          if( not WaitWithTimeout( $connhash ) )
          {
            die "Conn $conn: Query did not return as expected\n";
          }
        }
        if( scalar( @{ $connhash->{results} } ) == 0 )
        {
          die "Shouldn't happen\n";
        }
        print "Wait for connection $conn succeeded\n";
        HandleQueryResponse( $connhash, $connhash->{opts} );
      }
      else   # Can't run CheckPendingQueries() until all waits have been processed
      {
        # Checks that no queries have ended prematurely
        CheckPendingQueries();

        print "Connection $conn\n";

        ExecuteQuery( $connhash, $optshash, $query );
      }
    }

    print( ("-" x 80)."\n" );
    print @comments;
    @comments = ();

    my $opts;
    ( $conn, $opts, $query ) = /^(\d+)([a-z&]*)\s+(.*)/i;

    $optshash = { map { ($_ => 1) } split //, $opts };

    if( not defined $conn )
    {
      die "Malformed input";
    }
  }

# End of file, execute final query
  # Checks that no queries have ended prematurely
  CheckPendingQueries();

  print "Connection $conn\n";

  if( defined $query )
  {
    my $connhash = GetConnection( $conn );

    ExecuteQuery( $connhash, $optshash, $query );
  }

  foreach my $conn (@connections)
  {
    next if not defined $conn;

    next if not $conn->{pending};

    die "Connection $conn->{connnum}: Query still pending at end of test (test error)\n";
  }
  
  print(("=" x 80),"\n");
  print "Test completed successfully\n";
}

# Make connection to database
sub GetConnection
{
  my $conn = shift;

  my $connhash = $connections[ $conn ];

  if( not defined $connhash )
  {
    my $pgconn = Pg::connectdb( $PG_CONNECT_STRING );

    if( $pgconn->status != PGRES_CONNECTION_OK )
    {
      die "Connection to database failed: ".$pgconn->errorMessage
    }

    $connhash = $connections[ $conn ] = { db => $pgconn, results => [], pending => 0, connnum => $conn };
  }

  return $connhash;
}

# Execute query, asyncronously if requested.
sub ExecuteQuery
{
  my($connhash, $optshash, $query) = @_;

  return if not defined $query;

  my $vars;

#  print join( "|", @_ ),"\n";

  print "Query: $query\n";

  if( $query =~ /^(\w+(,\w+)*)=(.*)/ )   # Check for variables=form
  {
    $vars = $1;
    $query = $3;
    $optshash->{VARS} = 1;
  }

  if( $query =~ s/(\$(\w+))/defined( $GlobalVars{$2} )?$GlobalVars{$2}:$1/eg )
  {
    print "Substituted query: $query\n";
  }

  if( $connhash->{db}->sendQuery( $query ) != 1 )
  {
    die "PQsendquery failed: ".$connhash->{db}->errorMessage;
  }

  if( not defined $optshash->{'&'} )    # Not backgrounded
  {
    if( not WaitWithTimeout( $connhash ) )
    {
      die "Query did not return as expected\n";
    }

    my $res = HandleQueryResponse( $connhash, $optshash );

    # set variables if requested
    if( defined $vars )
    {
      AssignVars( $vars, $res );
    }
  }
  else
  {
    print "Query doesn't return immediatly\n";
    SetConnectionPending( $connhash, $optshash );
  }
}

# Make query on this connection as pending
sub SetConnectionPending
{
  my( $connhash, $opts ) = @_;

  $connhash->{pending} = 1;
  $connhash->{opts} = $opts;   # save options for later
}

# No queries should end during this call. This function is called just
# before a real query is sent, ie, after all wait commands have been
# processed.
sub CheckPendingQueries
{
  foreach my $conn (@connections)
  {
    next if not defined $conn;

    next if not $conn->{pending};

    if( CheckPendingInput( $conn ) )  # If done, it's not pending anymore
    {
      die "Connection $conn->{connnum}: Query ended prematurely\n";
    }
  }
}

# Processes the response to a query. It may be called directly after a query
# has been sent, or after a query has been waited on.
sub HandleQueryResponse
{
  my ($connhash,$optshash) = @_;

  my @vars;   # Values to return to assign to variables...

  if( defined $optshash->{'I'} )   # This is needed for initial drop table commands
  {
    print "Result ignored\n";
    $connhash->{results} = [];   # Clear results
    $connhash->{pending} = 0;    # No longer pending
    delete $connhash->{opts};
    return;
  }

  foreach my $res ( @{ $connhash->{results} } )  # May be multiple result sets
  {
    my $status = $res->resultStatus;

    my $goterror = ($status == PGRES_BAD_RESPONSE 
                 or $status == PGRES_NONFATAL_ERROR
                 or $status == PGRES_FATAL_ERROR );

    if( $goterror and defined $optshash->{'e'} )
    {
      print "Received expected error: ".$connhash->{db}->errorMessage;
      next;
    }
    if( $goterror and not defined $optshash->{'e'} )
    {
      die "Unexpected error: ".$connhash->{db}->errorMessage;
    }
    if( not $goterror and defined $optshash->{'e'} )
    {
      die "Expected error but didn't get one\n";
    }

    if( $status == PGRES_TUPLES_OK )
    {
      print "Query result:\n";
      $res->print( \*STDOUT, 1, 1, 0, 0, 0, 0, "|", "", "" );

      if( defined $optshash->{VARS} )    # Want to save a result
      {
        if( $res->ntuples < 1 )
        {
          die "Needed values to assign to variable(s)\n";
        }
        @vars = map { $res->getvalue(0,$_) } (0..$res->nfields-1);

        delete $optshash->{VARS};    # Got them
      }
    }
    else
    {
      my $status = $res->cmdStatus;
      # Strip OID from insert, otherwise the tests keep failing
      $status =~ s/(INSERT )(\d+ )(\d+)/$1$3/;
      print "Command result: $status\n";
    }
  }

  $connhash->{results} = [];   # Clear results
  $connhash->{pending} = 0;    # No longer pending
  delete $connhash->{opts};

  return \@vars;
}

# Checks connection for input
# Returns 1 if finished result is available
sub CheckPendingInput
{
  my $connhash = shift;

#  print "CheckPendingInput($connhash)\n";

  # Keep going until there's no more input to consume...
  while(1)
  {
    $connhash->{db}->consumeInput;
    return 0 if $connhash->{db}->isBusy;
    my $res = $connhash->{db}->getResult;

#    print "res=$connhash, status=".$res->resultStatus."\n";

    # When end of results, this is what we get...
    if( $res->resultStatus == PGRES_FATAL_ERROR )
    {
      # If it's the first response, save that
      if( scalar( @{ $connhash->{results} } ) == 0 )
      {
        push @{ $connhash->{results} }, $res;
        return 0;   # Not yet done
      }

      return 1;    # Finished
    }
    push @{ $connhash->{results} }, $res;   # save up results...
  }
}

# Waits for a query to finished, checking for input every RESOLUTION
sub WaitWithTimeout
{
  my $connhash = shift;

  for my $i (1.. TIMEOUT/RESOLUTION)
  {
    if( CheckPendingInput( $connhash ) )
    { return 1 }

    select(undef, undef, undef, RESOLUTION);   # Hires sleep, sortof :)
  }

  return 0;
}

# Assign vars
sub AssignVars
{
  my( $vars, $res ) = @_;

  my @varlist = split /,/,$vars;

  if( scalar(@varlist) != scalar(@$res) )
  {
    die "Variable assignment and result set don't match\n";
  }

  for my $var (0..$#varlist)
  {
    $GlobalVars{$varlist[$var]} = $res->[$var];
    print "Setting variable $varlist[$var] = '$res->[$var]'\n";
  }
}
